#include "Renderable.hpp"

bool Renderable::initialize(const RenderType& type)
{
    if (mIsInitialized)
        return false;

    glGenVertexArrays(1, &mVAO);
    glBindVertexArray(mVAO);
    glGenBuffers(1, &mVBO);
    glGenBuffers(1, &mEBO);

    if (type == RenderType::Axis)
    {
        std::array<float, 42> vertices = {
            0.0f,  0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 1.0f, // B
            0.0f,  0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 1.0f,

            0.0f,  0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 1.0f, // G
            0.0f,  1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 1.0f,

            0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, // R
            1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f,

        };

        std::array<uint32_t, 6> indices = { 5, 4, 3, 2, 1, 0 };

        glBindBuffer(GL_ARRAY_BUFFER, mVBO);
        glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices.data(), GL_STATIC_DRAW);

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mEBO);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(indices), indices.data(), GL_STATIC_DRAW);

        glEnable(GL_SMOOTH);
        glLineWidth(4.0f);

        mAmountOfIndices = static_cast<uint32_t>(indices.size());
    }
    else if (type == RenderType::Plane)
    {
        // Default is Plane
        std::array<float, 28> vertices = { 1.0f,  1.0f,  0.0f, 0.78f, 0.08f, 0.52f, 0.5f,
                                           1.0f,  -1.0f, 0.0f, 0.78f, 0.08f, 0.52f, 0.5f,
                                           -1.0f, -1.0f, 0.0f, 0.78f, 0.08f, 0.52f, 0.5f,
                                           -1.0f, 1.0f,  0.0f, 0.78f, 0.08f, 0.52f, 0.5f };
        std::array<uint32_t, 6> indices = {
            3, 1, 0, // first triangle
            3, 2, 1  // second triangle
        };

        glBindBuffer(GL_ARRAY_BUFFER, mVBO);
        glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices.data(), GL_STATIC_DRAW);

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mEBO);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(indices), indices.data(), GL_STATIC_DRAW);

        mAmountOfIndices = static_cast<uint32_t>(indices.size());
    }
    else if (type == RenderType::Point)
        glPointSize(3.0f);

    GLint positionAttribute = glGetAttribLocation(mRenderableShader.ID, "aPos");
    // position attribute
    glVertexAttribPointer(positionAttribute, 3, GL_FLOAT, GL_FALSE, 7 * sizeof(float), nullptr);
    glEnableVertexAttribArray(positionAttribute);

    // color attribute
    GLint colorAttribute = glGetAttribLocation(mRenderableShader.ID, "aColor");
    glVertexAttribPointer(colorAttribute, 4, GL_FLOAT, GL_FALSE, 7 * sizeof(float), (void*)(3 * sizeof(float)));
    glEnableVertexAttribArray(colorAttribute);

    glBindVertexArray(0);

    mType = type;
    mIsInitialized = true;

    return true;
}

Renderable::~Renderable()
{
    glDeleteVertexArrays(1, &mVAO);
    glDeleteBuffers(1, &mVBO);
    glDeleteBuffers(1, &mEBO);
}

void Renderable::setIndices(const std::vector<int32_t>& indices)
{
    glBindVertexArray(mVAO);

    mAmountOfIndices = static_cast<uint32_t>(indices.size());

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mEBO);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(int32_t), indices.data(), GL_STATIC_DRAW);

    glBindVertexArray(0);
}
void Renderable::setVertices(const std::vector<VIRNECT::Vertex<double>>& vertices, bool hasColor)
{
    glBindVertexArray(mVAO);

    std::vector<float> drawableVertices;

    for (const auto& vertice : vertices)
    {
        drawableVertices.push_back(static_cast<float>(vertice.mPos.x));
        drawableVertices.push_back(static_cast<float>(vertice.mPos.y));
        drawableVertices.push_back(static_cast<float>(vertice.mPos.z));

        // Color
        if (hasColor)
        {
            drawableVertices.push_back(static_cast<float>(vertice.mRGBA.x)); // R
            drawableVertices.push_back(static_cast<float>(vertice.mRGBA.y)); // G
            drawableVertices.push_back(static_cast<float>(vertice.mRGBA.z)); // B
            drawableVertices.push_back(static_cast<float>(vertice.mRGBA.w)); // A
        }
        else
        {
            drawableVertices.push_back(0.78f); // R
            drawableVertices.push_back(0.08f); // G
            drawableVertices.push_back(0.52f); // B
            drawableVertices.push_back(0.5f);  // A
        }
    }

    if (mType == RenderType::Point)
        mAmountOfIndices = static_cast<uint32_t>(drawableVertices.size());

    glBindBuffer(GL_ARRAY_BUFFER, mVBO);
    glBufferData(GL_ARRAY_BUFFER, drawableVertices.size() * sizeof(float), drawableVertices.data(), GL_STATIC_DRAW);

    GLint positionAttribute = glGetAttribLocation(mRenderableShader.ID, "aPos");
    // position attribute
    glVertexAttribPointer(positionAttribute, 3, GL_FLOAT, GL_FALSE, 7 * sizeof(float), nullptr);
    glEnableVertexAttribArray(positionAttribute);

    // color attribute
    GLint colorAttribute = glGetAttribLocation(mRenderableShader.ID, "aColor");
    glVertexAttribPointer(colorAttribute, 4, GL_FLOAT, GL_FALSE, 7 * sizeof(float), (void*)(3 * sizeof(float)));
    glEnableVertexAttribArray(colorAttribute);

    glBindVertexArray(0);
}

void Renderable::setVertices(const std::vector<VIRNECT::Vec3<double>>& vertices)
{
    glBindVertexArray(mVAO);

    std::vector<float> drawableVertices;

    for (const auto& vertice : vertices)
    {
        drawableVertices.push_back(static_cast<float>(vertice.x));
        drawableVertices.push_back(static_cast<float>(vertice.y));
        drawableVertices.push_back(static_cast<float>(vertice.z));

        if (mType == RenderType::Point)
        {
            // Color
            drawableVertices.push_back(1.0f); // R
            drawableVertices.push_back(0.0f); // G
            drawableVertices.push_back(0.0f); // B
            drawableVertices.push_back(1.0f); // A
        }
        else
        {
            // Color
            drawableVertices.push_back(0.78f); // R
            drawableVertices.push_back(0.08f); // G
            drawableVertices.push_back(0.52f); // B
            drawableVertices.push_back(0.5f);  // A
        }
    }

    if (mType == RenderType::Point)
        mAmountOfIndices = static_cast<uint32_t>(drawableVertices.size());

    glBindBuffer(GL_ARRAY_BUFFER, mVBO);
    glBufferData(GL_ARRAY_BUFFER, drawableVertices.size() * sizeof(float), drawableVertices.data(), GL_STATIC_DRAW);

    GLint positionAttribute = glGetAttribLocation(mRenderableShader.ID, "aPos");
    // position attribute
    glVertexAttribPointer(positionAttribute, 3, GL_FLOAT, GL_FALSE, 7 * sizeof(float), nullptr);
    glEnableVertexAttribArray(positionAttribute);

    // color attribute
    GLint colorAttribute = glGetAttribLocation(mRenderableShader.ID, "aColor");
    glVertexAttribPointer(colorAttribute, 4, GL_FLOAT, GL_FALSE, 7 * sizeof(float), (void*)(3 * sizeof(float)));
    glEnableVertexAttribArray(colorAttribute);

    glBindVertexArray(0);
}

void Renderable::setLocalScale(float x, float y, float z) const
{
    std::array<float, 16> scaleMatrix{};
    scaleMatrix.fill(0);

    scaleMatrix[0] = x;
    scaleMatrix[5] = y;
    scaleMatrix[10] = z;
    scaleMatrix[15] = 1.0f;

    mRenderableShader.use();
    uint32_t scaleLoc = glGetUniformLocation(mRenderableShader.ID, "scale");
    glUniformMatrix4fv(scaleLoc, 1, GL_FALSE, scaleMatrix.data());
}

void Renderable::setLocalPosition(float x, float y, float z) const
{
    std::array<float, 16> posMatrix{};
    posMatrix.fill(0);

    posMatrix[12] = x;
    posMatrix[13] = y;
    posMatrix[14] = z;

    posMatrix[0] = posMatrix[5] = posMatrix[10] = posMatrix[15] = 1.0f;

    mRenderableShader.use();
    uint32_t posLoc = glGetUniformLocation(mRenderableShader.ID, "locPos");
    glUniformMatrix4fv(posLoc, 1, GL_FALSE, posMatrix.data());
}

void Renderable::setPerspective(const VIRNECT::CameraCalibration& cameraCalibration)
{
    std::array<float, 16> projectionMatrix{};
    float farPlane = 10000.0f;
    float nearPlane = 0.0001f;

    projectionMatrix[0] = 2.0f * static_cast<float>(cameraCalibration.mFocalLength[0]) / static_cast<float>(cameraCalibration.mResolution[0]);
    projectionMatrix[5] = 2.0f * static_cast<float>(cameraCalibration.mFocalLength[1]) / static_cast<float>(cameraCalibration.mResolution[1]);
    projectionMatrix[8] = 1.0f - 2.0f * static_cast<float>(cameraCalibration.mPrincipalPoint[0]) / static_cast<float>(cameraCalibration.mResolution[0]);
    projectionMatrix[9] = 2.0f * static_cast<float>(cameraCalibration.mPrincipalPoint[1]) / static_cast<float>(cameraCalibration.mResolution[1]) - 1.0f;
    projectionMatrix[10] = (farPlane + nearPlane) / (nearPlane - farPlane);
    projectionMatrix[11] = -1.0f;
    projectionMatrix[14] = 2.0f * farPlane * nearPlane / (nearPlane - farPlane);

    mRenderableShader.use();

    uint32_t perspectiveLoc = glGetUniformLocation(mRenderableShader.ID, "perspective");
    glUniformMatrix4fv(perspectiveLoc, 1, GL_FALSE, projectionMatrix.data());
}

void Renderable::render(const VIRNECT::ObjectTrackResult& result) const
{
    if (result.mStatus == VIRNECT::TrackingState::NOT_TRACKED)
        return;

    std::array<float, 16> transform{};

    transform[0] = static_cast<float>(result.mRotationMatrix[0]);
    transform[4] = static_cast<float>(result.mRotationMatrix[1]);
    transform[8] = static_cast<float>(result.mRotationMatrix[2]);
    transform[12] = static_cast<float>(result.mTranslation.x);

    transform[1] = static_cast<float>(result.mRotationMatrix[3]);
    transform[5] = static_cast<float>(result.mRotationMatrix[4]);
    transform[9] = static_cast<float>(result.mRotationMatrix[5]);
    transform[13] = static_cast<float>(result.mTranslation.y);

    transform[2] = static_cast<float>(result.mRotationMatrix[6]);
    transform[6] = static_cast<float>(result.mRotationMatrix[7]);
    transform[10] = static_cast<float>(result.mRotationMatrix[8]);
    transform[14] = static_cast<float>(result.mTranslation.z);

    transform[15] = 1.0f;

    mRenderableShader.use();
    glBindVertexArray(mVAO);
    uint32_t transformLoc = glGetUniformLocation(mRenderableShader.ID, "transform");
    glUniformMatrix4fv(transformLoc, 1, GL_FALSE, transform.data());

    if (mType == RenderType::Axis)
        glDrawElements(GL_LINES, mAmountOfIndices, GL_UNSIGNED_INT, nullptr);
    else if (mType == RenderType::Point)
        glDrawArrays(GL_POINTS, 0, mAmountOfIndices);
    else
        glDrawElements(GL_TRIANGLES, mAmountOfIndices, GL_UNSIGNED_INT, nullptr);

    glBindVertexArray(0);
}
