// Copyright (C) 2020 VIRNECT CO., LTD.
// All rights reserved.

#pragma once
#include "Shader.hpp"

#include <VARAPI/TrackingFramework/TrackingFramework.hpp>
#include <array>
#include <glad/glad.h>
#include <iostream>

/**
 * Renderable class for rendering virtual object
 */
class Renderable
{
public:
    /**
     * Object's shape
     */
    enum class RenderType
    {
        Point, ///< Map points
        Model, ///< Model target
        Plane, ///< Plane which is attached to target
        Axis,  ///< Target's axis
        None   ///< None
    };

    /**
     * @brief Default initializer
     */
    Renderable() = default;

    /**
     * @brief Destruct instance in order to release GL objects
     */
    ~Renderable();

    /**
     * @brief Initialize Renderable with RenderType
     * @param type If place is set to initialize function, Renderable will render place shape
     * @return If initialized already, return false
     */
    bool initialize(const RenderType& type);

    /**
     * @brief Set Renderable vertices.
     * @param vertices Vertex struct of array for drawing
     * @param hasColor If flag set to false, default color will be set
     */
    void setVertices(const std::vector<VIRNECT::Vertex<double>>& vertices,bool hasColor = true);

    /**
     * @brief Set Renderable vertices.
     * @param vertices vertices for drawing
     */
    void setVertices(const std::vector<VIRNECT::Vec3<double>>& vertices);

    /**
     * @brief Set Renderable indicies.
     * @param indices indices for drawing
     */
    void setIndices(const std::vector<int32_t>& indices);

    /**
     * @brief Set Renderable scale. You can use target information of VIRNECT Track
     * @param x X scale
     * @param y Y scale
     * @param z Z scale
     */
    void setLocalScale(float x, float y, float z) const;

    /**
     * @brief Set Renderable local position.
     * @param x X position
     * @param y Y position
     * @param z Z position
     */
    void setLocalPosition(float x, float y, float z) const;

    /**
     * @brief Set Projection matrix using camera intrinsics parameter
     * @param cameraCalibration Camera intrinsic parameter
     */
    void setPerspective(const VIRNECT::CameraCalibration& cameraCalibration);

    /**
     * @brief Draw Renderable object
     * @param result VIRNECT ObjectTrackResult instance which has target's transformation information
     */
    void render(const VIRNECT::ObjectTrackResult& result) const;

private:
    const char* vertexShaderSource = "#version 330 core\n"
                                     "layout (location = 0) in vec3 aPos;\n"
                                     "layout (location = 1) in vec4 aColor;\n"
                                     "out vec4 ourColor;\n"
                                     "uniform mat4 transform;\n"
                                     "uniform mat4 perspective;\n"
                                     "uniform mat4 scale;\n"
                                     "uniform mat4 locPos;\n"
                                     "void main()\n"
                                     "{\n"
                                     "   ourColor = aColor;\n"
                                     "   gl_Position = perspective  * transform * locPos * scale * vec4(aPos, 1.0);\n"
                                     "}\0"; ///< Vertex shader code
    const char* fragmentShaderSource = "#version 330 core\n"
                                       "in vec4 ourColor;\n"
                                       "out vec4 FragColor;\n"
                                       "void main()\n"
                                       "{\n"
                                       "   FragColor = ourColor;\n"
                                       "}\n\0"; ///< Fragment shader code

    RenderType mType = RenderType::None;                                         ///< Renderable's type
    Shader mRenderableShader = Shader(vertexShaderSource, fragmentShaderSource); ///< Shader instance which has shader codes
    uint32_t mVBO = 0;                                                           ///< Vertex Buffer Object ID
    uint32_t mVAO = 0;                                                           ///< Vertex Array Object ID
    uint32_t mEBO = 0;                                                           ///< Element Buffer Object ID
    uint32_t mAmountOfIndices = 0;                                               ///< Drawable vertex size
    bool mIsInitialized = false;                                                 ///< Initialized flag
};
