#include "RenderingEngine.hpp"

int RenderingEngine::gPressedKey = -1;

void key_callback(GLFWwindow* window, int key, [[maybe_unused]] int scancode, int action, [[maybe_unused]] int mods)
{
    if (action == GLFW_PRESS)
        RenderingEngine::gPressedKey = key;
    else
        RenderingEngine::gPressedKey = -1;

    if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS)
        glfwSetWindowShouldClose(window, GLFW_TRUE);
}

RenderingEngine::RenderingEngine(uint32_t width, uint32_t height)
{
    // GL init
    if (glfwInit() == GLFW_FALSE)
        return;

    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_RESIZABLE, GL_FALSE);
    glfwWindowHint(GLFW_VISIBLE, GL_TRUE);

#if __APPLE__
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
#endif

    mWindow = glfwCreateWindow(width, height, "TRACK Sample", nullptr, nullptr);
    if (!mWindow)
    {
        glfwTerminate();
        return;
    }

    glfwSetKeyCallback(mWindow, key_callback);

    glfwMakeContextCurrent(mWindow);
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return;
    }
    glfwSwapInterval(1);

    mBackgroundTexture = std::make_unique<BackgroundTexture>(width, height);

    glViewport(0, 0, width, height);
    glEnable(GL_BLEND);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

RenderingEngine::~RenderingEngine()
{
    glfwDestroyWindow(mWindow);
    glfwTerminate();
}

void RenderingEngine::addTarget(const std::string& targetName, const VIRNECT::TargetInformation& target)
{
    glfwMakeContextCurrent(mWindow);

    auto& [axis, plane] = mTargetRenderable[targetName];
    axis.initialize(Renderable::RenderType::Axis);
    axis.setPerspective(mCalibration);
    axis.setLocalScale(static_cast<float>(target.mDimensions.mMax.x), static_cast<float>(target.mDimensions.mMax.y), static_cast<float>(target.mDimensions.mMax.x));
    axis.setLocalPosition(0, 0, 0);

    plane.initialize(Renderable::RenderType::Plane);
    plane.setPerspective(mCalibration);
    plane.setLocalScale(static_cast<float>(target.mDimensions.mMax.x), static_cast<float>(target.mDimensions.mMax.y), static_cast<float>(target.mDimensions.mMax.z));
    plane.setLocalPosition(0, 0, 0);
}

void RenderingEngine::addTarget(const std::string& targetName, const VIRNECT::CADModel& target)
{
    glfwMakeContextCurrent(mWindow);

    auto& [axis, model] = mTargetRenderable[targetName];

    axis.initialize(Renderable::RenderType::Axis);
    axis.setPerspective(mCalibration);
    axis.setLocalScale(0.05f, 0.05f, 0.05f);
    axis.setLocalPosition(0, 0, 0);

    model.initialize(Renderable::RenderType::Model);
    model.setPerspective(mCalibration);
    model.setLocalScale(1.0f, 1.0f, 1.0f);
    model.setLocalPosition(0, 0, 0);

    model.setVertices(target.mVertices, false);
    model.setIndices(target.mIndices);
}

bool RenderingEngine::updateMapPointsIfNeeded(uint64_t version, const std::vector<VIRNECT::Vec3<double>>& points, const std::string& targetName)
{
    if (mCurMapVersion[targetName] > version)
        return false;

    mCurMapVersion[targetName] = version;

    auto& [axis, point] = mTargetRenderable[targetName];

    if (axis.initialize(Renderable::RenderType::Axis))
    {
        axis.setPerspective(mCalibration);
        axis.setLocalScale(0.1f, 0.1f, 0.1f);
        axis.setLocalPosition(0, 0, 0);
    }

    if (point.initialize(Renderable::RenderType::Point))
    {
        point.setPerspective(mCalibration);
        point.setLocalScale(1, 1, 1);
        point.setLocalPosition(0, 0, 0);
    }

    point.setVertices(points);

    return true;
}

void RenderingEngine::render(const VIRNECT::ImageRGB& frame, const std::unordered_map<std::string, VIRNECT::ObjectTrackResult>& results)
{
    glfwMakeContextCurrent(mWindow);

    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    mBackgroundTexture->render(frame);

    for (const auto& [targetName, result] : results)
    {
        auto it = mTargetRenderable.find(targetName);
        if (it != mTargetRenderable.end())
        {
            auto& [axis, target] = it->second;

            target.render(result);
            axis.render(result);
        }
    }

    glfwSwapBuffers(mWindow);
    glfwPollEvents();
}
