// Copyright (C) 2020 VIRNECT CO., LTD.
// All rights reserved.

#pragma once

#include "BackgroundTexture.hpp"
#include "Renderable.hpp"

// clang-format off include GLAD before GLFW
#include <glad/glad.h>
// clang-format on
#include <GLFW/glfw3.h>
#include <memory>

/**
 * @brief Rendering Engine class
 */
class RenderingEngine
{
public:
    /**
     * @brief Initialize RenderingEngine with windows size
     * @param width window width
     * @param height window height
     */
    RenderingEngine(uint32_t width, uint32_t height);

    /**
     * @brief Destructor of RenderingEngine
     */
    ~RenderingEngine();

    /**
     * @brief Set camera intrinsics parameter to Engine for projection matrix
     * @param calibration Camera intrinsics parameter
     */
    void setCalibration(VIRNECT::CameraCalibration& calibration) { mCalibration = calibration; };

    /**
     * @brief Add target for adding virtual object
     * @param targetName Target name
     * @param target Plane target information
     */

    void addTarget(const std::string& targetName, const VIRNECT::TargetInformation& target);

    /**
     * @brief Add target for adding virtual object
     * @param targetName Target name
     * @param targets CAD Model
     */
    void addTarget(const std::string& targetName, const VIRNECT::CADModel& target);

    /**
     * @brief Add or Update map points
     * @param targetName Name of target
     * @param points Points from map data
     * @return Return true if map points are updated
     */
    bool updateMapPointsIfNeeded(uint64_t version, const std::vector<VIRNECT::Vec3<double>>& points, const std::string& targetName = "");

    /**
     * @brief Get map version
     * @return Version of current map points
     */
    uint64_t getMapVersion(const std::string& targetName) const { return mCurMapVersion.at(targetName); }

    /**
     * @brief Check window is closed by user input
     * @return If return true, window is closed
     */
    [[nodiscard]] bool isCloseEngine() const { return glfwWindowShouldClose(mWindow); }

    /**
     * @brief Draw Frame
     * @param frame Camera frame
     * @param results Target results from VIRNECT Track
     * @param points Map points from VIRNECT Track
     */
    void render(const VIRNECT::ImageRGB& frame, const std::unordered_map<std::string, VIRNECT::ObjectTrackResult>& results);

    static int gPressedKey; ///< Represents the key-code of the key that is currently pushed down. -1 if no key is pressed

private:
    GLFWwindow* mWindow = nullptr;                                                        ///< GLFW window
    VIRNECT::CameraCalibration mCalibration;                                              ///< Camera calibration
    std::unique_ptr<BackgroundTexture> mBackgroundTexture = nullptr;                      ///< Background texture object
    std::unordered_map<std::string, std::pair<Renderable, Renderable>> mTargetRenderable; ///< Renderable lists (target name, <axis,plane>)
    std::unordered_map<std::string, uint64_t> mCurMapVersion;                             ///< Current map versions
};
