// Copyright (C) 2020 VIRNECT CO., LTD.
// All rights reserved.

#include "RenderingEngine/RenderingEngine.hpp"

#include <VARAPI/TrackingFramework/Log.hpp>
#include <VARAPI/TrackingFramework/Tools.hpp>
#include <VARAPI/TrackingFramework/TrackingFramework.hpp>
#include <chrono>
#include <cstring>
#include <iostream>
#include <sstream>
#include <thread>

#define CT std::chrono::high_resolution_clock::now()

/**
 * @brief Callback function that gets invoked when the API logs a message
 * @param message Log message
 * @param level Log level
 */
void logCallback(const std::string& message, [[maybe_unused]] VIRNECT::LOG::LogLevel level)
{
    std::cout << message << std::endl;
}

/**
 * @brief CPP sample application to use the VIRNECT Track framework
 *
 * @param argc Command line argument count. A minimum of 7 command line arguments is required
 * @param argv Command line Arguments in following order:
 * 1) License key
 * 2) Path to target data (Folder containing TRACK_VARIMG_000)
 * 3) Target names separated with # (VARIMG_000#VARQR_000)
 * 4) Camera ID or path to image sequence
 * 5) Image Source as 'camera' or 'fileSequence'
 * 6) Path to camera calibration file ".json"
 * 7) Frames per second
 * 8) Optional special feature parameter: mapRecording
 * 9) If mapRecording is enabled, name of new map target
 * For example:
 * XXXX-XXXX-XXXX-XXXX ..\VARSDK\bin VARIMG_000#VARQR_000 fileSequence ..\seq\IMG\seq0000\P360 ..\calibration.json 60
 * XXXX-XXXX-XXXX-XXXX ..\VARSDK\bin VARIMG_000 fileSequence ..\seq\IMG\seq0000\P360 ..\calibration.json 60 mapRecording MAPTARGET
 * @return exit code of application
 */
int32_t main(int argc, char* argv[])
{
    std::cout << "\nVIRNECT Track - Sample\n" << std::endl;

    //----------------------------------------------------------
    // Define framework configuration parameters
    //
    uint32_t fps = 60;
    bool enableMapRecording = false;
    std::string mapTargetName = "";

    // Define CLI argument placeholders
    std::string licenseKey;
    std::string targetDataPath;
    std::vector<std::string> targetNames;
    std::string calibrationPath;
    int32_t cameraID = 0;
    std::string imageSequencePath;
    bool useCamera = true;

    //----------------------------------------------------------
    // Parse command line arguments into configuration
    //
    if (argc <= 7 || argc >= 11)
    {
        std::cout << "Usage:\n";
        std::cout << "    TRACK-Native.exe [licenseKey] [targetDataPath] [targetNames] [imageSource] [cameraID | fileSequencePath] [cameraCalibrationPath] [fps] [feature] [feature arguments]\n\n";
        std::cout << "Parameters:\n";
        std::cout << "    licenseKey                     Track license key\n";
        std::cout << "    targetDataPath                 Folder containing the target files\n";
        std::cout << "    targetNames                    List of Target names without \"TRACK_\" prefix, separated with \"#\"\n";
        std::cout << "    imageSource                    Define 'camera' or 'fileSequence' as an image source\n";
        std::cout << "    cameraID | fileSequencePath    USB camera ID as integer (starting with 0) or path to file sequence\n";
        std::cout << "    cameraCalibrationPath          Camera calibration JSON\n";
        std::cout << "    fps                            Frames per second\n";
        std::cout << "    feature                        Optional! Enable one of the following features:\n";
        std::cout << "     - 'mapRecording'              Enable Map recording mode:\n";
        std::cout << "                                     - will save recorded target in targetDataPath\n";
        std::cout << "                                     - requires an ImageTarget for initialization\n";
        std::cout << "                                     - requires additional parameter for the name of the recorded MAP target\n";
        std::cout << "\nExample:\n";
        std::cout << "    TRACK-Native.exe 0000-0000-0000-0000 ..\\targets\\ VARIMG_000#VARQR_000 camera 0  ..\\calibration.json 60\n\n";
        return -1;
    }

    licenseKey = argv[1];
    targetDataPath = argv[2];

    // Extract target names
    std::stringstream stringStream(argv[3]);
    std::string name;
    while (std::getline(stringStream, name, '#'))
        targetNames.push_back(name);

    // Determine playmode
    std::string imageSource = argv[4];

    if (imageSource == "camera")
        cameraID = std::stoi(argv[5]);
    else if (imageSource == "fileSequence")
    {
        imageSequencePath = argv[5];
        useCamera = false;
    }
    else
    {
        std::cout << "ERROR: Image Source '" << imageSource << "' does not equal 'camera' or 'fileSequence'" << std::endl;
        return -1;
    }

    // Set calibration
    calibrationPath = argv[6];

    // Convert parameter 6 to an integer
    fps = std::stoi(argv[7]);

    // Feature specific arguments
    if (argc >= 9)
    {
        if (std::string(argv[8]) == "mapRecording")
        {
            enableMapRecording = true;
            // Try to convert parameter 9 to an integer
            if (argc >= 10)
                mapTargetName = argv[9];
            else
            {
                std::cout << "Error: Additional argument for name of map target is missing" << std::endl;
                return -1;
            }
        }
        else
        {
            std::cout << "Error: Argument " << argv[8] << "could not be interpreted" << std::endl;
            return -1;
        }
    }
    //----------------------------------------------------------
    // Initialize Track framework
    //

    // Register callback function for logging
    VIRNECT::LOG::registerLogCallbackFunction(logCallback);

    // Set the license key
    if (!VIRNECT::LICENSE::setLicenseKey(licenseKey))
    {
        std::cout << "Error: license key could not be validated" << std::endl;
        return -1;
    }
    std::cout << "License key validated" << std::endl;

    // Use the calibration reader tool to load camera calibration from JSON file
    VIRNECT::CameraCalibration cameraCalibration;
    VIRNECT::TOOLS::readCalibrationFile(calibrationPath, cameraCalibration);

    // Instantiate tracking framework
    VIRNECT::TrackingFramework trackingFramework;

    // Setup framework standard configuration
    trackingFramework.setCameraCalibration(cameraCalibration);
    trackingFramework.setTargetDataPath(targetDataPath);
    trackingFramework.setTargetNames(targetNames);

    // Record MapTarget
    if (enableMapRecording)
        trackingFramework.enableMapTargetGeneration(true, mapTargetName, targetDataPath);

    // Initialize tracking framework
    if (!trackingFramework.initialize())
    {
        std::cout << "Error: Could not initialize framework. Aborting." << std::endl;
        return -1;
    }

    // Set image source
    if (useCamera)
        VIRNECT::TOOLS::activateImageSourceUSBCamera(cameraID);
    else
        VIRNECT::TOOLS::activateImageSourceFileSequence(imageSequencePath);

    //----------------------------------------------------------
    // Initialize rendering engine
    //
    RenderingEngine renderingEngine(cameraCalibration.mResolution[0], cameraCalibration.mResolution[1]);
    renderingEngine.setCalibration(cameraCalibration);

    // Set target information
    const auto infos = trackingFramework.getTargetInformation();

    std::vector<std::string> mapNames = {};

    for (const auto& [targetName, info] : infos)
    {
        switch (info.mType)
        {
        case VIRNECT::TargetType::MAP:
            mapNames.push_back(targetName);
            renderingEngine.updateMapPointsIfNeeded(trackingFramework.getMapVersion(targetName), trackingFramework.getMapPoints(targetName), targetName);
            break;

        case VIRNECT::TargetType::CADMODEL: renderingEngine.addTarget(targetName, trackingFramework.getCADModel()); break;

        default: renderingEngine.addTarget(targetName, info);
        }
    }

    if (enableMapRecording)
    {
        mapNames.push_back(mapTargetName);
        renderingEngine.updateMapPointsIfNeeded(trackingFramework.getMapVersion(mapTargetName), trackingFramework.getMapPoints(mapTargetName), mapTargetName);
    }

    // Calculate frame time in nanoseconds
    long frameTime = static_cast<long>(1E9 / fps);

    if (enableMapRecording)
        std::cout << "Map recording is active. It will start as soon as " << targetNames[0] << " is detected. Press S to save the target data file in " << targetDataPath << std::endl;

    //----------------------------------------------------------
    // Execute framework
    //
    bool processFrames = true;
    while (!renderingEngine.isCloseEngine() && processFrames)
    {
        // Run tracking framework
        auto frameworkStart = CT;
        processFrames = trackingFramework.process();
        auto frameworkFinish = CT;

        // Get tracking result
        const auto results = trackingFramework.getTrackingResult();

        for (const std::string& name : mapNames)
            if (trackingFramework.getMapVersion(name) > renderingEngine.getMapVersion(name))
                renderingEngine.updateMapPointsIfNeeded(trackingFramework.getMapVersion(name), trackingFramework.getMapPoints(name), name);

        // Render tracking result
        renderingEngine.render(trackingFramework.getLastProcessedImage(), results);

        // In case of active map recording, press "S" to save the target
        if (enableMapRecording && RenderingEngine::gPressedKey == GLFW_KEY_S)
        {
            if (trackingFramework.saveMapTarget())
                std::cout << "Map target '" << mapTargetName << "' successfully saved to " << targetDataPath << std::endl;
            else
                std::cout << "A problem occurred while saving the new map target" << std::endl;
        }

        // Control FPS
        long sleepTime = frameTime - (CT - frameworkStart).count();
        if (sleepTime > 0)
            std::this_thread::sleep_for(std::chrono::nanoseconds(sleepTime));
    }

    // Stop framework
    trackingFramework.stop();
}
